#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Base interface to the I/O stack running on the GECCO platform
Subsystems should be derived from the IOStack class.
"""

from __future__ import print_function
from builtins import range

import collections
import random
import socket
import struct
import enum
import json
import sys
import binascii
import time

default_address = "192.168.1.115"
default_port = 512
default_timeout = 0.3
default_retries = 3
default_verbosity = 0
default_max_packet_size = 300

SYS_IOSTACK = 0

class Command(object):
    """I/O stack command codes."""
    CMD_READ_REG = 0
    CMD_WRITE_REG = 1
    CMD_PING = 2
    CMD_APP_RESET = 3
    CMD_BOOT_RESET = 4
    CMD_FLASH_START = 5
    CMD_FLASH_WRITE_PAGE = 6
    CMD_FLASH_FINISH = 7
    CMD_CLEAR_BOOTPROT = 100
    CMD_FLASH_BOOT_START = 101
    CMD_FLASH_BOOT_WRITE_PAGE = 102
    CMD_FLASH_BOOT_FINISH = 103
    CMD_REPORT_ERR = 65525


class Register(object):
    """I/O stack registers."""
    REG_ETHERNET_CFG = 0
    REG_IDS = 1
    REG_BOOT_OPT = 2


class Status(object):
    """I/O stack error codes."""
    ERR_OKAY = 0
    ERR_UNKNOWN_SUBSYSTEM = 1
    ERR_UNKNOWN_COMMAND = 2
    ERR_INVALID_SIZE = 3
    ERR_INVALID_REGISTER = 4
    ERR_INVALID_MAC = 5
    ERR_UNHANDLED_ERROR = 6

class BootOptFlags(enum.IntFlag):
    AUTOSTART = 0x1
    CHECKCRC  = 0x2
    SETBOOTPROT = 0x4
    APPWATCHDOG = 0x8

# Generate lookup maps
for c in Register,Command, Status:
    c.lookup = {v: k for (k, v) in c.__dict__.items()
                if not k.startswith('__')}


class Error(Exception):
    """Base exception in case we need extended functionality in the future."""
    pass


class TimeoutError(Error):
    """Timeout occured while waiting for a response."""
    pass


class ResponseError(Error):
    """Received a malformed response."""
    pass


class RequestError(Error):
    """Reply indicated an invalid request."""
    pass


Response = collections.namedtuple("Response",
                                  "id subsystem_id response_code "
                                  "payload")

SubsystemResponse = collections.namedtuple("SubsystemResponse",
                                           "response_code payload")

IOStackConfig = collections.namedtuple("IOStackConfig",
                                       "ip subnet gateway mac")

IOStackIDs = collections.namedtuple("IOStackIDs",
                                       "serial EUI48 HWrev FWrev")

IOStackBootOpt = collections.namedtuple("IOStackBootOpt",
                                        "flags delay size CRC")

class IOStack(object):
    def __init__(self, ip, port=default_port, timeout=default_timeout,
                 max_retries=default_retries, verbosity=default_verbosity,
                 max_packet_size=default_max_packet_size,
                 interface_ip=None):
        """Connects to an I/O stack with the given address.

        Parameters
        ----------
        ip : string
            Destination address.
        port : int, optional
            Destination port (default: 512).
        timeout : float, optional
            Response timeout in seconds (default: 200 ms).
        max_retries : int, optional
            Maximum number of retries after a timeout (default: 3).
        verbosity : int, optional
            Verbosity level (default: 0, silent)
        max_packet_size : int, optional
            Expected maximum size of replies (default: 256 Bytes).
        interface_ip : str, optional
            IP address of local interface (default: let OS choose).
        """

        self.verbosity = verbosity
        self.max_packet_size = max_packet_size
        self.request_id = random.randint(0, 65535)
        self.max_retries = max_retries

        # Connect
        self.cs = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        if interface_ip is not None:
            # Bind to a specific interface
            for local_port in range(1025, 65535):
                try:
                    self.cs.bind((interface_ip, local_port))
                    if self.verbosity:
                        print("IOStack: bound to %s:%i" % (interface_ip, local_port), file=sys.stderr)
                    break
                except:
                    pass
            else:
                raise RuntimeError("error: no free UDP port on interface with IP %s" % interface_ip)

            self.cs.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.cs.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        self.cs.connect((ip, port))
        self.cs.settimeout(timeout)
        if self.verbosity:
            print("IOStack: opened connection to %s:%i, %.0f ms timeout" % (ip, port, 1000 * timeout), file=sys.stderr)

    def __del__(self):
        """Disconnects from an I/O stack"""
        self.cs.close()

    def set_timeout(self, timeout):
        self.cs.settimeout(timeout)

    def get_timeout(self):
        return self.cs.gettimeout()

    def multi_request(self, subsystem_id, request_code, payload, broadcast=False,
                      max_retries=None):
        """
        Send a request and yield one or more replies.

        Parameters
        ----------
        subsystem_id : int
            Identifier of the target subsystem.
        request_code : int
            Identifier of the command to execute.
        payload : bytearray or string
            Payload to append to the request.
        broadcast : bool, optional
            If true, wait for multiple answers (default: False).
        max_retries : int, optional
            Maximum number of retries after a timeout (None: use default
            value).

        Yields
        ------
        Response
            Responses with raw payload.
        """
        # Pack and transmit request
        self.request_id = (self.request_id + 1) % 65536
        request = struct.pack("<HBH", self.request_id, subsystem_id,
                              request_code) + payload

        self.cs.send(request)

        # Receive response
        trials_left = self.max_retries if max_retries is None else int(max_retries)
        while True:
            try:
                reply = self.cs.recv(self.max_packet_size)
            except socket.timeout:
                if trials_left > 0:  # Retry
                    trials_left -= 1
                    self.cs.send(request)
                    continue

                raise TimeoutError("network error: request timed out")

            if len(reply) < 5:
                raise ResponseError("response too short")

            header, payload = reply[:5], reply[5:]
            response = Response(*struct.unpack("<HBH", header),
                                payload=payload)

            if response.id != self.request_id:
                if self.verbosity:
                    print("skipping packet %s" % str(response), file=sys.stderr)

                continue

            # Perform sanity checks
            if response.subsystem_id == 0:
                if response.response_code not in Command.lookup:
                    raise ResponseError("invalid response code %i" % response.response_code)

                if response.response_code == Command.CMD_REPORT_ERR:
                    if len(response.payload) != 2:
                        raise ResponseError("invalid error size %i" % len(response.payload))

                    error_code, = struct.unpack("<H", response.payload)
                    if error_code not in Status.lookup:
                        raise ResponseError("invalid error code %i" % error_code)

                    raise RequestError(Status.lookup[error_code])

            if response.subsystem_id != subsystem_id:
                raise ResponseError("invalid subsystem id %i" % response.subsystem_id)

            yield response

            if not broadcast:
                break

    def request(self, subsystem_id, request_code, payload, max_retries=None):
        """Send a request and return the response.

        Parameters
        ----------
        subsystem_id : int
            Identifier of the target subsystem.
        request_code : int
            Identifier of the command to execute.
        payload : bytearray or string
            Payload to append to the request.
        max_retries : int, optional
            Maximum number of retries after a timeout (None: use default
            value).

        Returns
        -------
        Response
            Response with raw payload.
        """
        return list(self.multi_request(subsystem_id, request_code, payload,
                    False, max_retries=max_retries))[0]

    def multiread_register(self, register, type_code="<H", max_retries=None):
        """Read from a register from multiple devices and yield their contents.

        Parameters
        ----------
        register : int
            The register to read from.
        type_code : str
            The format of the response payload (see the documentation of Python's
            struct module).
        max_retries : int, optional
            Maximum number of retries after a timeout (None: use default
            value).

        Yields
        ------
        SubsystemResponse
            Response with payload decoded according to the type_code parameter.
        """
        payload = struct.pack("<H", register)

        for response in self.multi_request(SYS_IOSTACK, Command.CMD_READ_REG,
                                           payload, broadcast=True,
                                           max_retries=max_retries):
            yield SubsystemResponse(response.response_code,
                                    struct.unpack(type_code, response.payload))

    def read_register(self, register, type_code="<H", max_retries=None):
        """Reads from a register and returns its content.

        Parameters
        ----------
        register : int
            The register to read from.
        type_code : string
            The format of the response payload (see the documentation of Python's
            struct module).
        max_retries : int, optional
            Maximum number of retries after a timeout (None: use default
            value).

        Returns
        -------
        SubsystemResponse
            Response with payload decoded according to the type_code parameter.
        """
        payload = struct.pack("<H", register)

        response = self.request(SYS_IOSTACK, Command.CMD_READ_REG, payload,
                                max_retries=max_retries)

        return SubsystemResponse(response.response_code,
                                 struct.unpack(type_code, response.payload))

    def write_register(self, register, value, type_code="<H", max_retries=None):
        """Writes to a register.

        Parameters
        ----------
        register : int
            The register to write to.
        value : int, bytearray or str
            The value to write to the register. If typecode is not None, the
            value is first encoded with struct.pack using the given format.
        type_code : str
            The format of value (see the documentation of Python's struct
            module). If None, value is assumed to be a bytearray or string
            which will not be encoded.
        max_retries : int, optional
            Maximum number of retries after a timeout (None: use default
            value).
        """
        payload = struct.pack("<H", register)
        if type_code:  # pack value according to type code
            payload += struct.pack(type_code, *value)
        else:  # transparently pass value
            payload += value

        response = self.request(SYS_IOSTACK, Command.CMD_WRITE_REG, payload,
                                max_retries=max_retries)

        return response.response_code

    def ping(self, payload=None):
        """Probes the connection to the device by sending a random payload."""
        header_size = 5
        if payload is None:
            payload = bytearray(random.randint(0, 255)
                                for i in range(64 - header_size))
        elif len(payload) > self.max_packet_size - header_size:
            payload = payload[:self.max_packet_size - header_size]

        response = self.request(SYS_IOSTACK, Command.CMD_PING, payload, max_retries=0)
        if response.payload != payload:
            raise ResponseError("payload mismatch")
        return payload

    def reset(self):
        """Requests a SW reset of the device."""
        response = self.request(SYS_IOSTACK,Command.CMD_APP_RESET, b"")

    def boot_reset(self):
        """Requests a reset to the bootloader"""
        response = self.request(SYS_IOSTACK,Command.CMD_BOOT_RESET, b"")

    def get_config(self):
        """Reads the configuration of the device."""
        response = self.request(SYS_IOSTACK,Command.CMD_READ_REG,
                                struct.pack("<H",Register.REG_ETHERNET_CFG))
        config = IOStackConfig(".".join("%d" % m for m in bytes(response.payload[14:18])),
                               ".".join("%d" % m for m in bytes(response.payload[4:8])),
                               ".".join("%d" % m for m in bytes(response.payload[0:4])),
                               ":".join("%02X" % m for m in bytes(response.payload[8:14])))
        return config

    def set_config(self, ip=None, subnet=None, gateway=None, mac=None):
        """Writes the configuration of the device."""
        old_config=self.get_config()
        if ip is None:
            ip=old_config.ip
        if subnet is None:
            subnet=old_config.subnet
        if gateway is None:
            gateway=old_config.gateway
        if mac is None:
            mac=old_config.mac

        if len(ip.split('.')) != 4:
            raise Error("Invalid config")
        if len(subnet.split('.')) != 4:
            raise Error("Invalid config")
        if len(gateway.split('.')) != 4:
            raise Error("Invalid config")
        if len(mac.split(':')) != 6:
            raise Error("Invalid config")

        payload = bytes(map(int,gateway.split('.'))) + \
                  bytes(map(int,subnet.split('.'))) + \
                  bytes([int(x,16) for x in mac.split(':')]) + \
                  bytes(map(int,ip.split('.')))

        response = self.request(SYS_IOSTACK,Command.CMD_WRITE_REG,
                                struct.pack("<H",Register.REG_ETHERNET_CFG) + \
                                payload);

    def get_boot_opt(self):
        """Reads the boot options of the device."""
        response = self.request(SYS_IOSTACK,Command.CMD_READ_REG,
                                struct.pack("<H",Register.REG_BOOT_OPT))
        boot_opt = IOStackBootOpt(BootOptFlags(struct.unpack("<L",response.payload[0:4])[0]),
                                  struct.unpack("<L",response.payload[4:8])[0],
                                  struct.unpack("<L",response.payload[8:12])[0],
                                  "0x{:08X}".format(struct.unpack("<L",response.payload[12:16])[0]))
        return boot_opt

    def set_boot_opt(self,flags=None,delay=None):
        """Writes the boot options of the device."""
        old_boot_opt=self.get_boot_opt()
        if flags is None:
            flags=old_boot_opt.flags
        if delay is None:
            delay=old_boot_opt.delay

        payload = bytes(struct.pack("<L",int(flags))+
                        struct.pack("<L",int(delay))+
                        struct.pack("<L",int(old_boot_opt.size))+
                        struct.pack("<L",int(old_boot_opt.CRC,16)))
        response = self.request(SYS_IOSTACK,Command.CMD_WRITE_REG,
                                struct.pack("<H",Register.REG_BOOT_OPT) + \
                                payload);

    def get_IDs(self):
        """Reads the IDs of the device (serials, FW/HW revisions)."""
        response = self.request(SYS_IOSTACK,Command.CMD_READ_REG,
                                struct.pack("<H",Register.REG_IDS))
        config = IOStackIDs(":".join("%02X" % m for m in bytes(response.payload[0:16])),
                            ":".join("%02X" % m for m in bytes(response.payload[16:22])),
                            int(response.payload[22]),
                            response.payload[23:].decode())
        return config

    def clear_bootprot(self):
        """
        Clears the BOOTPROT so a new bootloader can be flashed
        A reboot is needed to be effective
        Also the SETBOOTPROT option should be disabled beforehand
        """
        response = self.request(SYS_IOSTACK,Command.CMD_CLEAR_BOOTPROT, b"")


    def upload_FW(self,fdata):
        """Uploads a new version of the FW into the system!"""
        padlen=((len(fdata)-1)//256+1)*256-(len(fdata))
        size=len(fdata)+padlen
        pad = b'\xFF' * padlen
        data=fdata+pad
        crc=0
        crc=binascii.crc32(data,crc)

        response = self.request(SYS_IOSTACK,Command.CMD_FLASH_START,
                                struct.pack("<LL",size,crc))

        print("Uploading FW with size {} bytes (padded) and CRC code 0x{:08X}".format(size,crc))

        for i in range((len(data) - 1) // 256 + 1):
            page = data[i * 256:(i + 1) * 256]
            print (".",end="",flush=True)
            header = struct.pack("<H",i)
            response = self.request(SYS_IOSTACK,Command.CMD_FLASH_WRITE_PAGE,
                                    header+page)

        print ("")
        print("Finishing FW Upload")
        response = self.request(SYS_IOSTACK,Command.CMD_FLASH_FINISH,b"")

    def upload_boot_FW(self,fdata):
        """
        Uploads a new version of the bootloader FW into the system!
        Only possible if running in the main Application FW (if it supports this feature)
        """
        padlen=((len(fdata)-1)//256+1)*256-(len(fdata))
        size=len(fdata)+padlen
        pad = b'\xFF' * padlen
        data=fdata+pad

        if size>65535:
            print("Bootloader FW too big. Aborting.")
            return

        response = self.request(SYS_IOSTACK,Command.CMD_FLASH_BOOT_START,
                                struct.pack("<L",size))

        print("Uploading Bootloader FW with size {} bytes (padded)".format(size))

        for i in range((len(data) - 1) // 256 + 1):
            page = data[i * 256:(i + 1) * 256]
            print (".",end="",flush=True)
            header = struct.pack("<H",i)
            response = self.request(SYS_IOSTACK,Command.CMD_FLASH_BOOT_WRITE_PAGE,
                                    header+page)

        print ("")
        print("Finishing FW Upload")
        response = self.request(SYS_IOSTACK,Command.CMD_FLASH_BOOT_FINISH,b"")

# allow to use this as an standalone test program
if __name__ == "__main__":
    import argparse

    def try_to_exec(expression,times=10):
        for attempt in range(times):
            try:
                exec(expression)
                break
            except:
                print (".",end="",flush=True)
                pass
            time.sleep(1)
        else:
            print (" Failed")
            return False
        print (" Success")
        return True

    # Parse command-line arguments
    arg_epilog="""Bootloader flags:
{}
""".format(json.dumps({i.name:i.value for i in BootOptFlags},indent=2))

    """
    {}
    Status Register Bits:
    {}
    ADC Channels:
    {}
    DAC Commands:
    {}
    DAC Channels:
    {}
    .format(json.dumps({i.name:i.value for i in Register},indent=2),
               json.dumps({i.name:i.value for i in StatusBits},indent=2),
               json.dumps({i.name:i.value for i in ADCAddress},indent=2),
               json.dumps({i.name:i.value for i in DACCommand},indent=2),
               json.dumps({i.name:i.value for i in DACAddress},indent=2),
               )
    """


    parser = argparse.ArgumentParser(
        description="Issue commands to the iostack subsystem of a GECCO board",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=arg_epilog)
    parser.add_argument("-i", "--ip", type=str, default=default_address,
                        help="IP address of the device (default: %s)" % default_address)
    parser.add_argument("-p","--port", type=int, default=default_port,
                        help="port (default: %i)" % default_port)
    parser.add_argument("-s","--string", type=str,
                        help="String used for pinging")
    parser.add_argument("-r","--reset", action='store_true',
                        help = "Request a reset")
    parser.add_argument("-b","--boot_reset", action='store_true',
                        help = "Request a reset to bootloader")
    parser.add_argument("-f","--force", action='store_true',
                        help = "Force command without trying to reset to bootloader")
    parser.add_argument("--get_config", action='store_true',
                        help = "Get target configuration")
    parser.add_argument("--get_boot_opt", action='store_true',
                        help = "Get target boot options")
    parser.add_argument("--get_IDs", action='store_true',
                        help = "Get target IDs")
    parser.add_argument("--ping", action='store_true',
                        help = "ping target")
    parser.add_argument("--set_ip", type=str,
                        help = "Set target IP address")
    parser.add_argument("--set_subnet", type=str,
                        help = "Set target subnet mask")
    parser.add_argument("--set_gateway", type=str,
                        help = "Set target gateway address")
    parser.add_argument("--set_boot_flags", type=str,
                        help = "Set bootloader flags")
    parser.add_argument("--set_boot_delay", type=int,
                        help = "Set bootloader delay (in ms, negative for infinite wait)")
    parser.add_argument("--fw_upload", type=str,
                        help = "Upload a new FW")
    parser.add_argument("--boot_fw_upload", type=str,
                        help = "Upload a new FW for the bootloader")
    parser.add_argument("--clear_bootprot", action='store_true',
                        help = "Clear the BOOTPROT protection of the bootloader")

    args = parser.parse_args()

    target=IOStack(ip=args.ip,port=args.port)

    if (args.fw_upload!=None):
        raw_data = open(args.fw_upload, 'rb').read()
        print("Requesting reset into bootloader")
        if (args.force or try_to_exec("target.boot_reset()")):
            print("Initiate FW upload")
            try_to_exec("target.upload_FW(raw_data)")

    if (args.boot_fw_upload!=None):
        raw_data = open(args.boot_fw_upload, 'rb').read()
        print("Initiate FW upload")
        try_to_exec("target.upload_boot_FW(raw_data)")

    if ((args.set_ip!=None) or
        (args.set_subnet!=None) or
        (args.set_gateway!=None)):
        if args.set_ip!=None:
            print ("Setting IP:{}".format(args.set_ip))
        if args.set_subnet!=None:
            print ("Setting subnet mask:{}".format(args.set_subnet))
        if args.set_gateway!=None:
            print ("Setting gateway:{}".format(args.set_gateway))
        input("Press Enter to confirm...")

        print("Requesting reset into bootloader")
        if (args.force or try_to_exec("target.boot_reset()")):
            print("Setting Ethernet configuration")
            try_to_exec("target.set_config(ip=args.set_ip,subnet=args.set_subnet,gateway=args.set_gateway)")
        exit()

    if ((args.set_boot_flags!=None) or
        (args.set_boot_delay!=None)):
        flags=args.set_boot_flags
        if args.set_boot_flags!=None:
            if args.set_boot_flags.isnumeric():
                flags=BootOptFlags(int(args.set_boot_flags))
            else:
                flags=BootOptFlags(0)
                for i in args.set_boot_flags.split("|"):
                    flags|=BootOptFlags[i]
            print ("Setting Bootloader flags:{}".format(flags))
        if args.set_boot_delay!=None:
            if (args.set_boot_delay<0):
                args.set_boot_delay= 0xFFFFFFFF
            print ("Setting Bootloader delay:{} ms".format(args.set_boot_delay))
        input("Press Enter to confirm...")

        print("Requesting reset into bootloader")
        if (args.force or try_to_exec("target.boot_reset()")):
            print("Setting Bootloader options")
            try_to_exec("target.set_boot_opt(flags=flags,delay=args.set_boot_delay)")
        exit()

    if args.reset:
        target.reset()

    if args.boot_reset:
        target.boot_reset()

    if args.clear_bootprot:
        target.clear_bootprot()

    if args.get_IDs:
        print(target.get_IDs())

    if args.get_config:
        print(target.get_config())

    if args.get_boot_opt:
        print(target.get_boot_opt())

    if args.ping:
        if (args.string!=None):
            print("Using the following string as payload: {}".format(args.string))
            print("Received the this reply from target:   {}".format(
                target.ping(bytes(args.string.encode('utf-8'))).decode('utf-8')))
        else:
            print("Using a random payload")
            target.ping(args.string)
            print("Done")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "phantom-hv"
copyright = "2024, Felix Werner"
author = "Felix Werner"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.linkcode",
]

autosummary_generate = True

templates_path = []
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = []
html_title = "phantom-hv"
html_theme = "furo"
html_theme_options = {
    "source_repository": "https://github.com/fwerner/phantom-hv/",
    "source_branch": "main",
    "source_directory": "docs/source/",
}

# -- Options for autodoc -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

# Automatically extract typehints when specified and place them in
# descriptions of the relevant function/method.
autodoc_typehints = "description"

# Don't show class signature with the class' name.
autodoc_class_signature = "separated"

# -- Options for sphinx.ext.linkcode -----------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/linkcode.html

def linkcode_resolve(domain, info):
    if domain != 'py' or not info['module']:
        return None

    root = "blob/main/src/"
    module = info['module']
    if module.count(".") < 2:
        filename = f"{module.replace('.', '/')}/__init__.py"
    else:
        filename = f"{module.replace('.', '/')}.py"

    return f"{html_theme_options['source_repository']}{root}{filename}"



#!/usr/bin/env python3
"""
Low-level control and monitoring command-line tool for Phantom HV modules.
"""

import argparse
import sys
import time

from phantomhv import PhantomHVIO, iostack


def _parse_set_dac_arg(arg):
    dac, level = map(int, arg.split(","))
    assert 0 <= dac <= 3
    assert 0 <= level <= 4095
    return dac, level


def main():
    parser = argparse.ArgumentParser("phantomhv-ctl", description=__doc__)
    subparsers = parser.add_subparsers(dest="command", required=True)
    subparsers.add_parser("boot", help="boot application slot")
    subparsers.add_parser("reset", help="reset into bootloader")
    flash_parser = subparsers.add_parser(
        "flash",
        help="flash binary firmware image into application slot",
    )
    flash_parser.add_argument(
        "-f",
        "--file",
        metavar="BIN_IMAGE",
        type=argparse.FileType("rb"),
        help="firmware image to write into flash",
    )
    subparsers.add_parser("unlock-hv", help="enable HV")
    subparsers.add_parser("lock-hv", help="disable HV")
    subparsers.add_parser(
        "enable-hv",
        help="enable HV channel",
    ).add_argument(
        "channel", choices=range(3), type=int, metavar="N", help="channel to enable"
    )
    subparsers.add_parser("disable-hv", help="disable HV channel").add_argument(
        "channel", choices=range(3), type=int, metavar="N", help="channel to disable"
    )
    set_parsers = subparsers.add_parser(
        "set", help="set hardware registers"
    ).add_subparsers(dest="set")
    set_dac_parser = set_parsers.add_parser("dac")
    set_dac_parser.add_argument(
        "dac-value",
        type=_parse_set_dac_arg,
        metavar="dac,level",
        help="set output level (0-4095) of DAC (0-3)",
    )
    monitor_parser = subparsers.add_parser(
        "monitor",
        help="continuously read and print slave states",
    )
    monitor_parser.add_argument(
        "--adcs-only",
        action="store_true",
        help="continuously read and print slave ADC readings only",
    )
    monitor_parser.add_argument(
        "--interval",
        type=float,
        default=0.5,
        metavar="dt",
        help="set monitoring interval (default: 0.5 s)",
    )

    for n, p in list(subparsers.choices.items()) + list(set_parsers.choices.items()):
        if n == "set":
            continue
        p.add_argument(
            "--ip",
            metavar="HOSTNAME",
            default=iostack.default_address,
            help="IP address or hostname of the device",
        )
        p.add_argument(
            "--port", default=iostack.default_port, help="Network port of the device"
        )
        p.add_argument(
            "--slot",
            type=int,
            default=0,
            choices=range(8),
            help="destination module (hardware slot)",
        )

    args = parser.parse_args()

    io = PhantomHVIO(args.ip)

    if args.command == "boot":
        io.boot_app(args.slot)
    elif args.command == "reset":
        io.reset(args.slot)
    elif args.command == "flash":
        print(f"Flashing {args.ip}...")
        state_before = io.read_slave_state(args.slot)
        io.flash_app(args.flash.read(), args.slot)
        state_after = io.read_slave_state(args.slot)
        if (
            state_after.spi_rx_errors != state_before.spi_rx_errors
            or state_after.spi_rx_overruns != state_before.spi_rx_overruns
        ):
            print("[ERROR] SPI communication error - please retry.")
            sys.exit(1)
    elif args.command == "unlock-hv":
        cfg = io.read_slave_dynamic_cfg(args.slot)
        cfg.hv_enable |= 1 << 3
        io.write_slave_dynamic_cfg(args.slot, cfg)
    elif args.command == "lock-hv":
        cfg = io.read_slave_dynamic_cfg(args.slot)
        cfg.hv_enable &= ~(1 << 3)
        io.write_slave_dynamic_cfg(args.slot, cfg)
    elif args.command == "disable-hv":
        cfg = io.read_slave_dynamic_cfg(args.slot)
        cfg.hv_enable &= ~(1 << args.channel)
        io.write_slave_dynamic_cfg(args.slot, cfg)
    elif args.command == "enable-hv":
        cfg = io.read_slave_dynamic_cfg(args.slot)
        cfg.hv_enable |= 1 << args.channel
        io.write_slave_dynamic_cfg(args.slot, cfg)
    elif args.command == "set" and args.set == "dac":
        dac, level = args.dac_value
        cfg = io.read_slave_dynamic_cfg(args.slot)
        cfg.hv_dac[dac] = level
        io.write_slave_dynamic_cfg(args.slot, cfg)

    if args.command == "monitor" and args.adcs_only:
        try:
            t0 = time.time()
            while True:
                t_before = time.time()
                adcs = io.read_slave_state(args.slot).adc_states
                t_after = time.time()
                print(
                    f"{t_after / 2 + t_before / 2:.3f} "
                    + " ".join(f"{adc:5.0f}" for adc in adcs)
                )
                if args.interval > 0:
                    t0 += args.interval
                    dt = t0 - time.time()
                    if dt > 0:
                        time.sleep(dt)
        except KeyboardInterrupt:
            print()
            pass
    elif args.command == "monitor":
        try:
            t0 = time.time()
            while True:
                print(io.read_slave_static_cfg(args.slot), end=" ")
                print(io.read_slave_dynamic_cfg(args.slot), end=" ")
                print(io.read_slave_state(args.slot), flush=True)

                t0 += args.interval
                dt = t0 - time.time()
                if dt > 0:
                    time.sleep(dt)
        except KeyboardInterrupt:
            print()
            pass



"""
Provides the class `PhantomHVIO` for low-level I/O.
"""

import struct
import threading
import time

from . import iostack


class PhantomHVSlaveStaticCfg:
    def __init__(self, buf):
        self.fw_rev, self.hw_rev, self.bl_rc, self.hw_uid = struct.unpack("<HHHQ", buf)

    def __str__(self):
        return str(self.__dict__)

    @staticmethod
    def size():
        return 14


class PhantomHVSlaveDynamicCfg:
    def __init__(self, buf):
        self.hv_enable, self.led_enable, self.pulse_enable = struct.unpack(
            "<" + "H" * 3, buf[:6]
        )
        self.hv_dac = list(struct.unpack("<" + "H" * 4, buf[6:14]))
        self.offset_dac = list(struct.unpack("<" + "H" * 4, buf[14:]))

    def pack(self):
        return struct.pack(
            "<" + "H" * 11,
            self.hv_enable,
            self.led_enable,
            self.pulse_enable,
            *self.hv_dac,
            *self.offset_dac
        )

    def __str__(self):
        return str(self.__dict__)

    @staticmethod
    def size():
        return 22


class PhantomHVSlaveState:
    def __init__(self, buf):
        self.status, self.spi_rx_errors, self.spi_rx_overruns, self.digital_states = (
            struct.unpack("<HHHH", buf[:8])
        )
        self.adc_states = list(struct.unpack("<" + "h" * 8, buf[8:24]))
        self.alerts, self.warnings, self.ticks = struct.unpack("<HHH", buf[24:])

    def __str__(self):
        return str(self.__dict__)

    @staticmethod
    def size():
        return 30


class PhantomHVIO(iostack.IOStack):
    PHANTOM_HV_MASTER_SUBSYSTEM = 0x11
    PHANTOM_HV_MASTER_CMD_SPI_IO_REQUEST = 0x0

    def __init__(
        self,
        ip,
        port=iostack.default_port,
        timeout=iostack.default_timeout,
        max_retries=iostack.default_retries,
        verbosity=iostack.default_verbosity,
        max_packet_size=iostack.default_max_packet_size,
        interface_ip=None,
    ):
        """Connects to the SPI I/O subsystem at the given address.

        Parameters
        ----------
        ip : string
            Destination address.
        port : int, optional
            Destination port (default: 512).
        timeout : float, optional
            Response timeout in seconds (default: 200 ms).
        verbosity : int, optional
            Verbosity level (default: 0, silent)
        max_packet_size : int, optional
            Expected maximum size of replies (default: 256 Bytes).
        interface_ip : str, optional
            IP address of local interface (default: let OS choose).
        """
        iostack.IOStack.__init__(
            self,
            ip=ip,
            port=port,
            timeout=timeout,
            max_retries=max_retries,
            verbosity=verbosity,
            max_packet_size=max_packet_size,
            interface_ip=interface_ip,
        )
        self.lock = threading.RLock()

    def _transfer(self, buffer):
        """Transmits the contents of buffer via SPI and returns the (same number of) bytes received."""
        with self.lock:
            return self.request(
                PhantomHVIO.PHANTOM_HV_MASTER_SUBSYSTEM,
                PhantomHVIO.PHANTOM_HV_MASTER_CMD_SPI_IO_REQUEST,
                buffer,
            )

    def _read_register(self, slot, register: int, Factory):
        assert register in (0, 1, 2)
        assert slot in range(8)
        with self.lock:
            response = self.request(
                PhantomHVIO.PHANTOM_HV_MASTER_SUBSYSTEM,
                PhantomHVIO.PHANTOM_HV_MASTER_CMD_SPI_IO_REQUEST,
                bytearray([slot, (1 << 4) | register, 0] + [0] * Factory.size()),
            )
        assert len(response.payload) == Factory.size() + 2
        return Factory(response.payload[2:])

    def read_slave_static_cfg(self, slot):
        assert slot in range(8)
        with self.lock:
            return self._read_register(slot, 0, PhantomHVSlaveStaticCfg)

    def read_slave_dynamic_cfg(self, slot):
        assert slot in range(8)
        with self.lock:
            return self._read_register(slot, 1, PhantomHVSlaveDynamicCfg)

    def read_slave_state(self, slot):
        assert slot in range(8)
        with self.lock:
            return self._read_register(slot, 2, PhantomHVSlaveState)

    def write_slave_dynamic_cfg(self, slot, cfg):
        assert slot in range(8)
        with self.lock:
            self.request(
                PhantomHVIO.PHANTOM_HV_MASTER_SUBSYSTEM,
                PhantomHVIO.PHANTOM_HV_MASTER_CMD_SPI_IO_REQUEST,
                bytearray([slot, 0x21]) + cfg.pack(),
            )

    def _write_page(self, page: int, data: str, slot: int, min_timeout=0.5):
        assert 0 <= slot <= 8
        assert len(data) <= 256
        assert page >= 0 and page < 4096
        header = struct.pack("<BBB", slot, (3 << 4) | (page >> 8) & 0xF, page & 0xFF)
        pad = b"\x00" * (256 - len(data))

        with self.lock:
            # Increase timeout if needed
            old_timeout = self.get_timeout()
            if old_timeout < min_timeout:
                self.set_timeout(min_timeout)

            self.request(
                PhantomHVIO.PHANTOM_HV_MASTER_SUBSYSTEM,
                PhantomHVIO.PHANTOM_HV_MASTER_CMD_SPI_IO_REQUEST,
                header + data + pad,
            )

            # Restore old timeout
            if old_timeout < min_timeout:
                self.set_timeout(old_timeout)

    def flash_app(self, data: str, slot: int, sleep_interval=0.1):
        print("Flashing", end="", flush=True)
        with self.lock:
            for i in range((len(data) - 1) // 256 + 1):
                print(".", end="", flush=True)
                page = data[i * 256 : (i + 1) * 256]
                self._write_page(i, page, slot)
                if sleep_interval > 0:
                    time.sleep(sleep_interval)
            print()

    def boot_app(self, slot: int):
        with self.lock:
            self.request(
                PhantomHVIO.PHANTOM_HV_MASTER_SUBSYSTEM,
                PhantomHVIO.PHANTOM_HV_MASTER_CMD_SPI_IO_REQUEST,
                bytearray([slot, 0x50]),
            )

    def reset(self, slot: int):
        with self.lock:
            self.request(
                PhantomHVIO.PHANTOM_HV_MASTER_SUBSYSTEM,
                PhantomHVIO.PHANTOM_HV_MASTER_CMD_SPI_IO_REQUEST,
                bytearray([slot, 0x60]),
            )



"""
Provides `PhantomHVStateBuffer` which buffers register reads.
"""

import time

from .io import PhantomHVIO
from . import iostack


class PhantomHVStateBuffer:
    """Caches all registers of a Phantom HV crate for quick repeated access via
    the `slots` member (which contains `num_slots` `PhantomHVSlotStateBuffer`
    objects). Call `update()` after initialisation to populate the cache."""

    def __init__(self, ip, port=iostack.default_port, num_slots=1):
        self.io = PhantomHVIO(ip, port=port)
        self.num_slots = num_slots
        self.slot = [
            PhantomHVSlotStateBuffer(self, self.io, slot) for slot in range(num_slots)
        ]
        self.static_cfgs = []

    def update(self, full=False):
        """Reads all slave registers and then atomically updates the internal
        states. Any exceptions such as iostack.TimeoutError or OSError are
        passed through.

        By default (full equals False), the static configurations will only be
        read once."""

        static_cfgs, dynamic_cfgs, states = [], [], []
        for slot in range(self.num_slots):
            if slot > len(self.static_cfgs) - 1 or full:
                static_cfgs.append(self.io.read_slave_static_cfg(slot))
            else:
                static_cfgs.append(self.static_cfgs[slot])
            dynamic_cfgs.append(self.io.read_slave_dynamic_cfg(slot))
            states.append(self.io.read_slave_state(slot))

        self.last_update = time.time()
        for slot in range(self.num_slots):
            self.slot[slot].update_from(
                self.last_update, static_cfgs[slot], dynamic_cfgs[slot], states[slot]
            )

        self.static_cfgs = static_cfgs


class PhantomHVSlotStateBuffer:
    """Caches all registers of a Phantom HV module for quick repeated access via
    the `channels` member (which contains `num_channels`
    `PhantomHVChannelStateBuffer` objects). Call `update()` after initialisation
    to populate the cache.

    Setters issue synchronous requests to the hardware."""

    def __init__(self, parent, io, slot, num_channels=3):
        self.parent = parent
        self.io = io
        self.slot = slot
        self.num_channels = num_channels
        self.channel = [
            PhantomHVChannelStateBuffer(self, io, self.slot, channel)
            for channel in range(num_channels)
        ]

    def update(self, full=False):
        if not hasattr(self, "static_cfg") or full:
            static_cfg = self.io.read_slave_static_cfg(self.slot)
        else:
            static_cfg = self.static_cfg
        dynamic_cfg = self.io.read_slave_dynamic_cfg(self.slot)
        state = self.io.read_slave_state(self.slot)
        self.update_from(time.time(), static_cfg, dynamic_cfg, state)
        self.parent.update()

    def update_from(self, update_time, static_cfg, dynamic_cfg, state):
        self.last_update = update_time
        self.static_cfg = static_cfg
        self.dynamic_cfg = dynamic_cfg
        self.state = state
        for channel in self.channel:
            channel.update_from(self.last_update, static_cfg, dynamic_cfg, state)

    @property
    def hv_unlocked_ext(self):
        return bool(self.state.digital_states & 0x4)

    @property
    def hv_unlocked(self):
        return bool(self.state.digital_states & 0x1)

    @hv_unlocked.setter
    def hv_unlocked(self, unlock):
        if unlock and (self.state.digital_states & 0x4) == 0:
            raise ValueError("HV interlock error")

        cfg = self.dynamic_cfg
        if unlock:
            cfg.hv_enable |= 1 << 3
        else:
            cfg.hv_enable = 0  # disable all HV supplies at the same time to avoid alert
        self.io.write_slave_dynamic_cfg(self.slot, cfg)

    @property
    def hvs(self):
        return [self.state.adc_states[channel] * 2000 / 16384 for channel in range(3)]


class PhantomHVChannelStateBuffer:
    """Caches all registers of a Phantom HV channel for quick repeated access.

    Setters issue synchronous requests to the hardware."""

    def __init__(self, parent, io, slot, channel):
        self.parent = parent
        self.io = io
        self.slot = slot
        self.channel = channel

    def update(self):
        self.parent.update()

    def update_from(self, update_time, static_cfg, dynamic_cfg, state):
        self.last_update = update_time
        self.static_cfg = static_cfg
        self.dynamic_cfg = dynamic_cfg
        self.state = state

    @property
    def hv_enabled(self):
        return bool(self.dynamic_cfg.hv_enable & (1 << self.channel))

    @hv_enabled.setter
    def hv_enabled(self, enable):
        cfg = self.dynamic_cfg
        if enable:
            cfg.hv_enable |= 1 << self.channel
        else:
            cfg.hv_enable &= ~(1 << self.channel)
        self.io.write_slave_dynamic_cfg(self.slot, cfg)

    @property
    def hv(self) -> float:
        return self.state.adc_states[self.channel] * 2000 / 16384

    @property
    def hv_set(self):
        return self.dynamic_cfg.hv_dac[self.channel] / 2.048

    @hv_set.setter
    def hv_set(self, hv: float):
        if hv is None:
            hv = 0.0
        hv = max(0.0, min(hv, 2000.0))
        cfg = self.dynamic_cfg
        cfg.hv_dac[self.channel] = round(hv * 2.048)
        self.io.write_slave_dynamic_cfg(self.slot, cfg)

    @property
    def current(self):
        return (
            self.state.adc_states[3 + self.channel]
            * (2.5 / 16384)
            / (330e-3 * (50 / 2))
        )



"""
Control and monitoring tools for the Phantom HV supply developed for SWGO at
MPIK.

Use the command-line tools `phantomhv-ctl` or `phantomhv-webui`, or the I/O
classes `PhantomHVIO` or `PhantomHVStateBuffer` to communicate with the
hardware.
"""

__all__ = [
    "PhantomHVIO",
    "PhantomHVStateBuffer",
]

from .io import PhantomHVIO
from .state_buffer import PhantomHVStateBuffer



#!/usr/bin/env python3
"""
Web UI to monitor and control a Phantom HV crate.
"""

import argparse

from phantomhv import iostack
from phantomhv.webui_components import PhantomHVWebUI


def parse_host_port(s, default_host=None, default_port=None):
    if ":" in s:
        if s.startswith(":"):
            return default_host, int(s[1:])
        else:
            host, port = s.split(":")
            return host, int(port)

    return s, default_port


def main():
    parser = argparse.ArgumentParser("phantomhv-webui", description=__doc__)
    parser.add_argument(
        "--address",
        default=f"{iostack.default_address}:{iostack.default_port}",
        required=False,
        type=lambda s: parse_host_port(s, default_port=iostack.default_port),
        metavar="ip:port",
        help=f"IP address or hostname of the device (default: {iostack.default_address}:{iostack.default_port})",
    )
    parser.add_argument(
        "--num-slots",
        default=1,
        type=int,
        choices=range(1, 9),
        help="number of modules to display (default: 1)",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--show",
        default=False,
        action="store_true",
        help="open UI in native window",
    )
    group.add_argument(
        "--bind",
        default=("127.0.0.1", 8080),
        type=lambda s: parse_host_port(s, default_host="127.0.0.1", default_port=8080),
        metavar="hostname:port",
        help="bind web server to a specific network interface/port (default: 127.0.0.1:8080; use 0.0.0.0 to bind to all interfaces)",
    )

    args = parser.parse_args()
    hv_host, hv_port = args.address
    bind_host, bind_port = args.bind

    web_ui = PhantomHVWebUI(hv_host, hv_port, num_slots=args.num_slots)

    try:
        web_ui.run(show=args.show, bind_host=bind_host, bind_port=bind_port)
    except KeyboardInterrupt:
        print()
        print("caught ctrl+c; shutting down")



"""
Components for the Web UI.
"""

import collections
import importlib
import time
import types

from nicegui import app, ui, run, elements
import numpy as np
import plotly.graph_objects as go

from phantomhv import iostack, state_buffer


def get_hv_setter(
    channel_state: state_buffer.PhantomHVChannelStateBuffer,
    channel_hv_set: elements.number.Number,
):
    """Returns a function that sets the channel's HV to `channel_hv_set`'s
    value.

    The function then retrieves the actual DAC value that has been programmed
    and updates the spinbox accordingly.
    """

    def set_hv():
        channel_state.hv_set = channel_hv_set.value
        channel_state.update()
        channel_hv_set.value = round(channel_state.hv_set)

    return set_hv


def get_hv_timeseries_updater(
    slot: state_buffer.PhantomHVSlotStateBuffer, duration_sec=65
):
    """Returns a function that on each call puts the latest voltage measurements
    of the given module slot into a ring buffer of `duration_sec` seconds
    length, and returns these measurements as a tuple of (numpy datetime64
    timestamps, voltages 2d array)."""
    t, v = collections.deque(), collections.deque()

    def get_hv_timeseries():
        time_ms = round(slot.last_update * 1000)
        voltages = slot.hvs

        if len(t) == 0 or time_ms > t[-1]:
            t.append(time_ms)
            v.append(voltages)

            while (t[-1] - t[0]) / 1000 > duration_sec:
                t.popleft()
                v.popleft()

        return np.array(t, dtype="datetime64[ms]"), np.array(v)

    return get_hv_timeseries


class PhantomHVWebUI:
    _header_font_classes = "font-medium text-purple-600"

    def __init__(
        self, host, port=iostack.default_port, num_slots=1, update_interval=0.2
    ):
        self.host = host
        self.port = port
        self.num_slots = num_slots
        self.default_update_interval = update_interval
        self.timeout_notification = None
        self.update_time = None

        self.hv_state = state_buffer.PhantomHVStateBuffer(
            host, port=port, num_slots=num_slots
        )
        self.hv_timeseries = []

    def run(self, show=False, bind_host=None, bind_port=None):
        self.init_ui()

        run_args = {
            "title": "Phantom HV Web UI",
            "dark": True,
            "reload": False,
            "show": show,
            "native": show,
        }

        if show:
            run_args["window_size"] = (1200, 130 + 160 * self.num_slots)
        else:
            run_args["host"] = bind_host
            run_args["port"] = bind_port

        ui.run(**run_args)

    def on_exception(self, error):
        ui.notify(error)

    def init_ui(self):
        ui.query("body").style(
            "font-family: 'Source Code Pro', 'Source Code Variable', Courier, monospace"
        ).classes("text-lg")

        ui.switch.default_props("color=deep-purple")
        ui.label(f"Phantom HV Web UI [{self.host}:{self.port}]").classes(
            self._header_font_classes
        )

        app.add_static_files(
            "/woff2", importlib.resources.files("phantomhv") / "resources" / "woff2"
        )
        ui.add_head_html(
            r"""
        <style>
        @font-face{
            font-family: 'Source Code Variable';
            font-weight: 200 900;
            font-style: normal;
            font-stretch: normal;
            font-display: swap;
            src: url('/woff2/SourceCodeVF-Upright.otf.woff2') format('woff2');
        }
        </style>
        """
        )
        app.on_exception(self.on_exception)
        ui.timer(0, self.update_state, once=True)

    def build_ui(self):
        with ui.grid(columns=9 * "auto " + "1fr").classes(
            "w-full items-baseline text-center gap-y-0"
        ):
            # Setup header row
            ui.label.default_classes(self._header_font_classes)
            ui.label("Slot")
            ui.label("Interlock")
            ui.label("HV unlock")
            ui.label("Channel")
            ui.label("HV enable")
            ui.label("Set voltage").classes("col-span-2")
            ui.label("Voltage")
            ui.label("Current")
            ui.label("History").classes("text-left")
            ui.label.default_classes(remove=self._header_font_classes)

            for slot in range(self.num_slots):
                slot_state = self.hv_state.slot[slot]
                for channel in range(3):
                    # Setup one row per channel
                    channel_state = slot_state.channel[channel]
                    if channel == 0:
                        # Slot (spans 3 rows)
                        ui.label(f"{slot}").classes("row-span-3")

                        # Interlock (spans 3 rows)
                        ui.label("ok").classes(
                            "row-span-3 text-emerald-500 font-medium"
                        ).bind_visibility_from(slot_state, "hv_unlocked_ext")
                        ui.label("error").classes(
                            "row-span-3 text-rose-500 font-medium"
                        ).bind_visibility_from(
                            slot_state, "hv_unlocked_ext", lambda b: not b
                        )

                        # HV unlock (cannot span multiple rows because of alignment issues)
                        # We can only toggle HV unlock if the interlock is 'ok'.
                        slot_hv_enable = (
                            ui.switch()
                            .bind_value(slot_state, "hv_unlocked")
                            .bind_enabled_from(slot_state, "hv_unlocked_ext")
                        ).classes("place-self-center")
                    else:
                        # HV unlock
                        ui.label("")

                    # Channel
                    ui.label(f"{channel}")

                    # HV enable
                    channel_hv_enable = (
                        ui.switch()
                        .bind_value(channel_state, "hv_enabled")
                        .bind_enabled_from(slot_hv_enable, "value")
                        .classes("place-self-center")
                    )

                    # Set voltage
                    channel_hv_set = (
                        ui.number(
                            "",
                            min=0,
                            max=2000,
                            precision=0,
                            value=round(channel_state.hv_set),
                        )
                        .bind_enabled_from(channel_hv_enable, "value")
                        .props("dense suffix='V' size='xl'")
                    )
                    ui.button(
                        "set",
                        color="deep-purple",
                        on_click=get_hv_setter(channel_state, channel_hv_set),
                    ).bind_enabled_from(channel_hv_enable, "value")

                    # Voltage
                    ui.label().bind_text_from(
                        channel_state,
                        "hv",
                        backward=lambda hv: f"{hv:.1f} V",
                    ).classes("text-right")

                    # Current
                    ui.label().bind_text_from(
                        channel_state,
                        "current",
                        backward=lambda i: f"{1000 * i:.1f} mA",
                    ).classes("text-right")

                    if channel == 0:
                        # History (spans 3 rows)
                        timeseries = TimeSeriesPlot(
                            yaxis_title="Voltage / V",
                        )
                        timeseries.plot.classes("row-span-3 w-full h-40 self-center")
                        self.hv_timeseries.append(
                            (
                                timeseries,
                                get_hv_timeseries_updater(slot_state),
                            )
                        )

    async def update_state(self):
        """Attempts to update the Phantom HV state buffer. On first success,
        builds the UI. On failure, displays a notification and 'connecting'
        message."""

        try:
            await run.io_bound(self.hv_state.update)
        except (iostack.TimeoutError, OSError):
            self.num_failures += 1
            reconnect_interval = self.default_update_interval * 2 ** min(
                6, self.num_failures
            )
            if (
                self.timeout_notification is None
                or self.timeout_notification.visible == False
            ):
                ui.notify(
                    "Lost connection to Phantom HV - will periodically attempt to reconnect..."
                )
            if self.timeout_notification is None:
                with ui.page_sticky("top-right", 8, 8) as self.timeout_notification:
                    with ui.row().classes("items-center bg-purple-900 p-2 rounded"):
                        ui.spinner(color="text-current")
                        ui.label("Connecting...")
            else:
                self.timeout_notification.set_visibility(True)

            ui.timer(
                reconnect_interval,
                self.update_state,
                once=True,
            )
            return

        now = time.time()
        self.num_failures = 0
        if self.timeout_notification is not None:
            self.timeout_notification.set_visibility(False)

        if self.update_time is None:
            self.build_ui()  # it's the first time we've succeeded in connecting: build the UI
            self.update_time = now + self.default_update_interval
        else:
            self.update_time += self.default_update_interval
            while self.update_time < now:
                self.update_time += self.default_update_interval

        for timeseries, update in self.hv_timeseries:
            x, y = update()
            timeseries.update(x, y)

        ui.timer(
            self.update_time - now,
            self.update_state,
            once=True,
        )


def to_plotly_json_with_config(self):
    json = self.to_dict()
    json["config"] = {"staticPlot": True}
    return json


class TimeSeriesPlot:
    """Plotly-based time series plot for NiceGUI."""

    def __init__(
        self,
        xaxis_title=None,
        yaxis_title=None,
        xaxis_tickangle=0,
    ):
        self.fig = go.Figure()
        self.fig.update_layout(
            template="plotly_dark",
            font_family="'Source Code Pro', 'Source Code Variable', Courier, monospace",
            uirevision=True,
            datarevision=True,
            margin=dict(l=0, r=0, t=0, b=0),
            xaxis=dict(
                title=xaxis_title,
                tickangle=xaxis_tickangle,
                automargin="bottom",
            ),
            yaxis=dict(title=yaxis_title, tickformat="~r"),
        )
        self.plot = ui.plotly(self.fig)
        self.fig.to_plotly_json = types.MethodType(to_plotly_json_with_config, self.fig)

    def update(self, x, y):
        if self.fig.data:
            for i, _y in enumerate(y.T):
                self.fig.data[i]["x"] = x
                self.fig.data[i]["y"] = _y
        else:
            for i, _y in enumerate(y.T):
                self.fig.add_trace(go.Scatter(x=x, y=_y, name=str(i), hoverinfo="none"))

        self.plot.update()
